@@ -14,6 +14,8 @@
 #    under the License.
 import collections
 import json
+import base64
+from contextlib import closing
 
 import eventlet
 from oslo_config import cfg
@@ -218,6 +220,32 @@ class LXDDriverTest(test.NoDBTestCase):
 
         self.assertEqual(['mock-instance-1', 'mock-instance-2'], instances)
 
+    @mock.patch('nova.virt.lxd.driver.IMAGE_API')
+    @mock.patch('nova.virt.lxd.driver.lockutils.lock')
+    def test_spawn_unified_image(self, lock, IMAGE_API=None):
+        def image_get(*args, **kwargs):
+            raise lxdcore_exceptions.LXDAPIException(MockResponse(404))
+        self.client.images.get_by_alias.side_effect = image_get
+        self.client.images.exists.return_value = False
+        image = {'name': mock.Mock(), 'disk_format': 'raw'}
+        IMAGE_API.get.return_value = image
+
+        def download_unified(*args, **kwargs):
+            # unified image with metadata
+            # structure is gzipped tarball, content:
+            # /
+            #  metadata.yaml
+            #  rootfs/
+            unified_tgz = 'H4sIALpegVkAA+3SQQ7CIBCFYY7CCXRAppwHo66sTVpYeHsh0a'\
+                          'Ru1A2Lxv/bDGQmYZLHeM7plHLa3dN4NX1INQyhVRdV1vXFuIML'\
+                          '4lVVopF28cZKp33elCWn2VpTjuWWy4e5L/2NmqcpX5Z91zdawD'\
+                          'HqT/kHrf/E+Xo0Vrtu9fTn+QMAAAAAAAAAAAAAAADYrgfk/3zn'\
+                          'ACgAAA=='
+            with closing(open(kwargs['dest_path'], 'wb+')) as img:
+                img.write(base64.b64decode(unified_tgz))
+        IMAGE_API.download = download_unified
+        self.test_spawn()
+
     @mock.patch('nova.virt.configdrive.required_by')
     def test_spawn(self, configdrive, neutron_failure=None):
         def container_get(*args, **kwargs):
@@ -616,8 +644,8 @@ class LXDDriverTest(test.NoDBTestCase):
     def test_attach_interface(self):
         expected = {
             'hwaddr': '00:11:22:33:44:55',
-            'parent': 'qbr0123456789a',
-            'nictype': 'bridged',
+            'parent': 'tin0123456789a',
+            'nictype': 'physical',
             'type': 'nic',
         }
 
@@ -645,7 +673,8 @@ class LXDDriverTest(test.NoDBTestCase):
             'type': network_model.VIF_TYPE_OVS,
             'address': '00:11:22:33:44:55',
             'network': {
-                'bridge': 'fakebr'}}
+                'bridge': 'fakebr'},
+            'devname': 'tap0123456789a'}
 
         lxd_driver = driver.LXDDriver(None)
         lxd_driver.init_host(None)
@@ -653,17 +682,52 @@ class LXDDriverTest(test.NoDBTestCase):
 
         lxd_driver.attach_interface(ctx, instance, image_meta, vif)
 
-        self.assertTrue('eth1' in profile.devices)
-        self.assertEqual(expected, profile.devices['eth1'])
+        self.assertTrue('tap0123456789a' in profile.devices)
+        self.assertEqual(expected, profile.devices['tap0123456789a'])
+        profile.save.assert_called_once_with(wait=True)
+
+    def test_detach_interface_legacy(self):
+        profile = mock.Mock()
+        profile.devices = {
+            'eth0': {
+                'nictype': 'bridged',
+                'parent': 'lxdbr0',
+                'hwaddr': '00:11:22:33:44:55',
+                'type': 'nic'
+            },
+            'root': {
+                'path': '/',
+                'type': 'disk'
+            },
+        }
+        self.client.profiles.get.return_value = profile
+
+        ctx = context.get_admin_context()
+        instance = fake_instance.fake_instance_obj(
+            ctx, name='test', memory_mb=0)
+        vif = {
+            'id': '0123456789abcdef',
+            'type': network_model.VIF_TYPE_OVS,
+            'address': '00:11:22:33:44:55',
+            'network': {
+                'bridge': 'fakebr'}}
+
+        lxd_driver = driver.LXDDriver(None)
+        lxd_driver.init_host(None)
+
+        lxd_driver.detach_interface(ctx, instance, vif)
+
+        self.vif_driver.unplug.assert_called_once_with(
+            instance, vif)
+        self.assertEqual(['root'], sorted(profile.devices.keys()))
         profile.save.assert_called_once_with(wait=True)
 
     def test_detach_interface(self):
         profile = mock.Mock()
         profile.devices = {
-            'eth0': {
-                'name': 'eth0',
-                'nictype': 'bridged',
-                'parent': 'lxdbr0',
+            'tap0123456789a': {
+                'nictype': 'physical',
+                'parent': 'tin0123456789a',
                 'hwaddr': '00:11:22:33:44:55',
                 'type': 'nic'
             },
@@ -758,16 +822,18 @@ class LXDDriverTest(test.NoDBTestCase):
             auth=True)
         mountpoint = '/dev/sdd'
 
+        driver.brick_get_connector = mock.MagicMock()
+        driver.brick_get_connector_properties = mock.MagicMock()
         lxd_driver = driver.LXDDriver(None)
         lxd_driver.init_host(None)
-
-        lxd_driver.storage_driver.connect_volume = mock.MagicMock()
+        # driver.brick_get_connector = mock.MagicMock()
+        # lxd_driver.storage_driver.connect_volume = mock.MagicMock()
         lxd_driver.attach_volume(
             ctx, connection_info, instance, mountpoint, None, None, None)
 
         lxd_driver.client.profiles.get.assert_called_once_with(instance.name)
-        lxd_driver.storage_driver.connect_volume.assert_called_once_with(
-            connection_info['data'])
+        # driver.brick_get_connector.connect_volume.assert_called_once_with(
+        #     connection_info['data'])
         profile.save.assert_called_once_with()
 
     def test_detach_volume(self):
@@ -815,7 +881,8 @@ class LXDDriverTest(test.NoDBTestCase):
         lxd_driver = driver.LXDDriver(None)
         lxd_driver.init_host(None)
 
-        lxd_driver.storage_driver.disconnect_volume = mock.MagicMock()
+        driver.brick_get_connector = mock.MagicMock()
+        driver.brick_get_connector_properties = mock.MagicMock()
         lxd_driver.detach_volume(connection_info, instance, mountpoint, None)
 
         lxd_driver.client.profiles.get.assert_called_once_with(instance.name)
@@ -883,6 +950,23 @@ class LXDDriverTest(test.NoDBTestCase):
         self.client.containers.get.assert_called_once_with(instance.name)
         container.unfreeze.assert_called_once_with(wait=True)
 
+    def test_resume_state_on_host_boot(self):
+        container = mock.Mock()
+        state = mock.Mock()
+        state.memory = dict({'usage': 0, 'usage_peak': 0})
+        state.status_code = 102
+        container.state.return_value = state
+        self.client.containers.get.return_value = container
+        ctx = context.get_admin_context()
+        instance = fake_instance.fake_instance_obj(
+            ctx, name='test', memory_mb=0)
+
+        lxd_driver = driver.LXDDriver(None)
+        lxd_driver.init_host(None)
+
+        lxd_driver.resume_state_on_host_boot(ctx, instance, None, None)
+        container.start.assert_called_once_with(wait=True)
+
     def test_rescue(self):
         profile = mock.Mock()
         profile.devices = {
@@ -1038,8 +1122,84 @@ class LXDDriverTest(test.NoDBTestCase):
                                 '\n'),
             meminfo,
         ]
-
+        lxd_config = {
+            'environment': {
+                'storage': 'dir',
+            },
+            'config': {}
+        }
         lxd_driver = driver.LXDDriver(None)
+        lxd_driver.client = mock.MagicMock()
+        lxd_driver.client.host_info = lxd_config
+        value = lxd_driver.get_available_resource(None)
+        # This is funky, but json strings make for fragile tests.
+        value['cpu_info'] = json.loads(value['cpu_info'])
+
+        self.assertEqual(expected, value)
+
+    @mock.patch('socket.gethostname', mock.Mock(return_value='fake_hostname'))
+    @mock.patch('nova.virt.lxd.driver.open')
+    @mock.patch.object(driver.utils, 'execute')
+    def test_get_available_resource_zfs(self, execute, open):
+        expected = {
+            'cpu_info': {
+                "features": "fake flag goes here",
+                "model": "Fake CPU",
+                "topology": {"sockets": "10", "threads": "4", "cores": "5"},
+                "arch": "x86_64", "vendor": "FakeVendor"
+            },
+            'hypervisor_hostname': 'fake_hostname',
+            'hypervisor_type': 'lxd',
+            'hypervisor_version': '011',
+            'local_gb': 2222,
+            'local_gb_used': 200,
+            'memory_mb': 10000,
+            'memory_mb_used': 8000,
+            'numa_topology': None,
+            'supported_instances': [
+                ('i686', 'lxd', 'exe'),
+                ('x86_64', 'lxd', 'exe'),
+                ('i686', 'lxc', 'exe'),
+                ('x86_64', 'lxc', 'exe')],
+            'vcpus': 200,
+            'vcpus_used': 0}
+
+        execute.side_effect = [
+            ('Model name:          Fake CPU\n'
+             'Vendor ID:           FakeVendor\n'
+             'Socket(s):           10\n'
+             'Core(s) per socket:  5\n'
+             'Thread(s) per core:  4\n\n',
+             None),
+            ('2.17T\n', None),
+            ('200.4G\n', None),
+            ('1.8T\n', None)
+        ]
+
+        meminfo = mock.MagicMock()
+        meminfo.__enter__.return_value = six.moves.cStringIO(
+            'MemTotal: 10240000 kB\n'
+            'MemFree:   2000000 kB\n'
+            'Buffers:     24000 kB\n'
+            'Cached:      24000 kB\n')
+
+        open.side_effect = [
+            six.moves.cStringIO('flags: fake flag goes here\n'
+                                'processor: 2\n'
+                                '\n'),
+            meminfo,
+        ]
+        lxd_config = {
+            'environment': {
+                'storage': 'zfs',
+            },
+            'config': {
+                'storage.zfs_pool_name': 'lxd',
+            }
+        }
+        lxd_driver = driver.LXDDriver(None)
+        lxd_driver.client = mock.MagicMock()
+        lxd_driver.client.host_info = lxd_config
         value = lxd_driver.get_available_resource(None)
         # This is funky, but json strings make for fragile tests.
         value['cpu_info'] = json.loads(value['cpu_info'])
