@@ -15,6 +15,7 @@
 
 from __future__ import absolute_import
 
+import errno
 import io
 import json
 import os
@@ -24,10 +25,13 @@ import shutil
 import socket
 import tarfile
 import tempfile
+import hashlib
 
 import eventlet
 import nova.conf
 import nova.context
+from contextlib import closing
+
 from nova import exception
 from nova import i18n
 from nova import image
@@ -62,11 +66,10 @@ import psutil
 from oslo_concurrency import lockutils
 from nova.compute import task_states
 from oslo_utils import excutils
+from oslo_utils import strutils
 from nova.virt import firewall
 
 _ = i18n._
-_LW = i18n._LW
-_LE = i18n._LE
 
 lxd_opts = [
     cfg.StrOpt('root_dir',
@@ -96,9 +99,33 @@ BASE_DIR = os.path.join(
     CONF.instances_path, CONF.image_cache_subdirectory_name)
 
 
+def _last_bytes(file_like_object, num):
+    """Return num bytes from the end of the file, and remaning byte count.
+
+    :param file_like_object: The file to read
+    :param num: The number of bytes to return
+
+    :returns: (data, remaining)
+    """
+
+    try:
+        file_like_object.seek(-num, os.SEEK_END)
+    except IOError as e:
+        # seek() fails with EINVAL when trying to go before the start of
+        # the file. It means that num is larger than the file size, so
+        # just go to the start.
+        if e.errno == errno.EINVAL:
+            file_like_object.seek(0, os.SEEK_SET)
+        else:
+            raise
+
+    remaining = file_like_object.tell()
+    return (file_like_object.read(), remaining)
+
+
 def _neutron_failed_callback(event_name, instance):
-    LOG.error(_LE('Neutron Reported failure on event '
-                  '%(event)s for instance %(uuid)s'),
+    LOG.error('Neutron Reported failure on event '
+              '%(event)s for instance %(uuid)s',
               {'event': event_name, 'uuid': instance.name},
               instance=instance)
     if CONF.vif_plugging_is_fatal:
@@ -165,6 +192,28 @@ def _get_fs_info(path):
             'used': used}
 
 
+def _get_zpool_info(pool):
+    """Get free/used/total disk space in a zfs pool."""
+    def _get_zpool_attribute(attribute):
+        value, err = utils.execute('zpool', 'list',
+                                   '-o', attribute,
+                                   '-H', pool,
+                                   run_as_root=True)
+        if err:
+            msg = _('Unable to parse zpool output.')
+            raise exception.NovaException(msg)
+        value = strutils.string_to_bytes('{}B'.format(value.strip()),
+                                         return_int=True)
+        return value
+
+    total = _get_zpool_attribute('size')
+    used = _get_zpool_attribute('alloc')
+    available = _get_zpool_attribute('free')
+    return {'total': total,
+            'available': available,
+            'used': used}
+
+
 def _get_power_state(lxd_state):
     """Take a lxd state code and translate it to nova power state."""
     state_map = [
@@ -194,6 +243,17 @@ def _sync_glance_image_to_lxd(client, context, image_ref):
             lock_path, external=True,
             lock_file_prefix='lxd-image-{}'.format(image_ref)):
 
+        # NOTE(jamespage): Re-query by image_ref to ensure
+        #                  that another process did not
+        #                  sneak infront of this one and create
+        #                  the same image already.
+        try:
+            client.images.get_by_alias(image_ref)
+            return
+        except lxd_exceptions.LXDAPIException as e:
+            if e.response.status_code != 404:
+                raise
+
         try:
             image_file = tempfile.mkstemp()[1]
             manifest_file = tempfile.mkstemp()[1]
@@ -204,26 +264,92 @@ def _sync_glance_image_to_lxd(client, context, image_ref):
                     image_id=image_ref, reason=_('Bad image format'))
             IMAGE_API.download(context, image_ref, dest_path=image_file)
 
-            metadata = {
-                'architecture': image.get(
-                    'hw_architecture', obj_fields.Architecture.from_host()),
-                'creation_date': int(os.stat(image_file).st_ctime)}
-            metadata_yaml = json.dumps(
-                metadata, sort_keys=True, indent=4,
-                separators=(',', ': '),
-                ensure_ascii=False).encode('utf-8') + b"\n"
+            # It is possible that LXD already have the same image
+            # but NOT aliased as result of previous publish/export operation
+            # (snapshot from openstack).
+            # In that case attempt to add it again
+            # (implicitly via instance launch from affected image) will produce
+            # LXD error - "Image with same fingerprint already exists".
+            # Error does not have unique identifier to handle it we calculate
+            # fingerprint of image as LXD do it and check if LXD already have
+            # image with such fingerprint.
+            # If any we will add alias to this image and will not re-import it
+            def add_alias():
 
-            tarball = tarfile.open(manifest_file, "w:gz")
-            tarinfo = tarfile.TarInfo(name='metadata.yaml')
-            tarinfo.size = len(metadata_yaml)
-            tarball.addfile(tarinfo, io.BytesIO(metadata_yaml))
-            tarball.close()
+                def lxdimage_fingerprint():
+                    def sha256_file():
+                        sha256 = hashlib.sha256()
+                        with closing(open(image_file, 'rb')) as f:
+                            for block in iter(lambda: f.read(65536), b''):
+                                sha256.update(block)
+                        return sha256.hexdigest()
 
-            with open(manifest_file, 'rb') as manifest:
+                    return sha256_file()
+
+                fingerprint = lxdimage_fingerprint()
+                if client.images.exists(fingerprint):
+                    LOG.info(
+                        'Image with fingerprint %(fingerprint)s already exists'
+                        'but not accessible by alias %(alias)s, add alias',
+                        {'fingerprint': fingerprint, 'alias': image_ref})
+                    lxdimage = client.images.get(fingerprint)
+                    lxdimage.add_alias(image_ref, '')
+                    return True
+
+                return False
+
+            if add_alias():
+                return
+
+            # up2date LXD publish/export operations produce images which
+            # already contains /rootfs and metdata.yaml in exported file.
+            # We should not pass metdata explicitly in that case as imported
+            # image will be unusable bacause LXD will think that it containts
+            # rootfs and will not extract embedded /rootfs properly.
+            # Try to detect if image content already has metadata and not pass
+            # explicit metadata in that case
+            def imagefile_has_metadata(image_file):
+                try:
+                    with closing(tarfile.TarFile.open(
+                        name=image_file, mode='r:*')) as tf:
+                        try:
+                            tf.getmember('metadata.yaml')
+                            return True
+                        except KeyError:
+                            pass
+                except tarfile.ReadError:
+                    pass
+                return False
+
+            if imagefile_has_metadata(image_file):
+                LOG.info('Image %(alias)s already has metadata, '
+                         'skipping metadata injection...',
+                         {'alias': image_ref})
                 with open(image_file, 'rb') as image:
-                    image = client.images.create(
-                        image.read(), metadata=manifest.read(),
-                        wait=True)
+                    image = client.images.create(image.read(), wait=True)
+            else:
+                metadata = {
+                    'architecture': image.get(
+                        'hw_architecture',
+                        obj_fields.Architecture.from_host()),
+                    'creation_date': int(os.stat(image_file).st_ctime)}
+                metadata_yaml = json.dumps(
+                    metadata, sort_keys=True, indent=4,
+                    separators=(',', ': '),
+                    ensure_ascii=False).encode('utf-8') + b"\n"
+
+                tarball = tarfile.open(manifest_file, "w:gz")
+                tarinfo = tarfile.TarInfo(name='metadata.yaml')
+                tarinfo.size = len(metadata_yaml)
+                tarball.addfile(tarinfo, io.BytesIO(metadata_yaml))
+                tarball.close()
+
+                with open(manifest_file, 'rb') as manifest:
+                    with open(image_file, 'rb') as image:
+                        image = client.images.create(
+                            image.read(), metadata=manifest.read(),
+                            wait=True)
+
             image.add_alias(image_ref, '')
 
         finally:
@@ -231,6 +357,43 @@ def _sync_glance_image_to_lxd(client, context, image_ref):
             os.unlink(manifest_file)
 
 
+def brick_get_connector_properties(multipath=False, enforce_multipath=False):
+    """Wrapper to automatically set root_helper in brick calls.
+    :param multipath: A boolean indicating whether the connector can
+                      support multipath.
+    :param enforce_multipath: If True, it raises exception when multipath=True
+                              is specified but multipathd is not running.
+                              If False, it falls back to multipath=False
+                              when multipathd is not running.
+    """
+
+    root_helper = utils.get_root_helper()
+    return connector.get_connector_properties(root_helper,
+                                              CONF.my_ip,
+                                              multipath,
+                                              enforce_multipath)
+
+
+def brick_get_connector(protocol, driver=None,
+                        use_multipath=False,
+                        device_scan_attempts=3,
+                        *args, **kwargs):
+    """Wrapper to get a brick connector object.
+    This automatically populates the required protocol as well
+    as the root_helper needed to execute commands.
+    """
+
+    root_helper = utils.get_root_helper()
+    if protocol.upper() == "RBD":
+        kwargs['do_local_attach'] = True
+    return connector.InitiatorConnector.factory(
+        protocol, root_helper,
+        driver=driver,
+        use_multipath=use_multipath,
+        device_scan_attempts=device_scan_attempts,
+        *args, **kwargs)
+
+
 class LXDLiveMigrateData(migrate_data.LiveMigrateData):
     """LiveMigrateData for LXD."""
 
@@ -263,12 +426,6 @@ class LXDDriver(driver.ComputeDriver):
         self.firewall_driver = firewall.load_driver(
             default='nova.virt.firewall.NoopFirewallDriver')
 
-        self.storage_driver = connector.InitiatorConnector.factory(
-            'ISCSI', utils.get_root_helper(),
-            use_multipath=CONF.libvirt.volume_use_multipath,
-            device_scan_attempts=CONF.libvirt.num_iscsi_scan_tries,
-            transport='default')
-
     def init_host(self, host):
         """Initialize the driver on the host.
 
@@ -303,7 +460,10 @@ class LXDDriver(driver.ComputeDriver):
 
     def get_info(self, instance):
         """Return an InstanceInfo object for the instance."""
-        container = self.client.containers.get(instance.name)
+        try:
+            container = self.client.containers.get(instance.name)
+        except lxd_exceptions.NotFound:
+            raise exception.InstanceNotFound(instance_id=instance.uuid)
 
         state = container.state()
         mem_kb = state.memory['usage'] >> 10
@@ -368,8 +528,8 @@ class LXDDriver(driver.ComputeDriver):
                         error_callback=_neutron_failed_callback):
                     self.plug_vifs(instance, network_info)
             except eventlet.timeout.Timeout:
-                LOG.warn(_LW('Timeout waiting for vif plugging callback for '
-                             'instance %(uuid)s'), {'uuid': instance['name']})
+                LOG.warn('Timeout waiting for vif plugging callback for '
+                         'instance %(uuid)s', {'uuid': instance['name']})
                 if CONF.vif_plugging_is_fatal:
                     self.destroy(
                         context, instance, network_info, block_device_info)
@@ -456,8 +616,8 @@ class LXDDriver(driver.ComputeDriver):
             container.delete(wait=True)
         except lxd_exceptions.LXDAPIException as e:
             if e.response.status_code == 404:
-                LOG.warning(_LW('Failed to delete instance. '
-                                'Container does not exist for %(instance)s.'),
+                LOG.warning('Failed to delete instance. '
+                            'Container does not exist for %(instance)s.',
                             {'instance': instance.name})
             else:
                 raise
@@ -492,8 +652,8 @@ class LXDDriver(driver.ComputeDriver):
             self.client.profiles.get(instance.name).delete()
         except lxd_exceptions.LXDAPIException as e:
             if e.response.status_code == 404:
-                LOG.warning(_LW('Failed to delete instance. '
-                                'Profile does not exist for %(instance)s.'),
+                LOG.warning('Failed to delete instance. '
+                            'Profile does not exist for %(instance)s.',
                             {'instance': instance.name})
             else:
                 raise
@@ -528,7 +688,7 @@ class LXDDriver(driver.ComputeDriver):
         utils.execute(
             'chmod', '755', instance_attrs.container_path, run_as_root=True)
         with open(console_path, 'rb') as f:
-            log_data, _ = utils.last_bytes(f, MAX_CONSOLE_BYTES)
+            log_data, _ = _last_bytes(f, MAX_CONSOLE_BYTES)
             return log_data
 
     def get_host_ip_addr(self):
@@ -551,8 +711,9 @@ class LXDDriver(driver.ComputeDriver):
         more information/
         """
         profile = self.client.profiles.get(instance.name)
-
-        device_info = self.storage_driver.connect_volume(
+        protocol = connection_info['driver_volume_type']
+        storage_driver = brick_get_connector(protocol)
+        device_info = storage_driver.connect_volume(
             connection_info['data'])
         disk = os.stat(os.path.realpath(device_info['path']))
         vol_id = connection_info['data']['volume_id']
@@ -588,7 +749,9 @@ class LXDDriver(driver.ComputeDriver):
             del profile.devices[vol_id]
             profile.save()
 
-        self.storage_driver.disconnect_volume(connection_info['data'], None)
+        protocol = connection_info['driver_volume_type']
+        storage_driver = brick_get_connector(protocol)
+        storage_driver.disconnect_volume(connection_info['data'], None)
 
     def attach_interface(self, context, instance, image_meta, vif):
         self.vif_driver.plug(instance, vif)
@@ -596,45 +759,41 @@ class LXDDriver(driver.ComputeDriver):
 
         profile = self.client.profiles.get(instance.name)
 
-        interfaces = []
-        for key, val in profile.devices.items():
-            if key.startswith('eth'):
-                interfaces.append(key)
-        net_device = 'eth{}'.format(len(interfaces))
-
-        network_config = lxd_vif.get_config(vif)
-        if 'bridge' in network_config:
-            config_update = {
-                net_device: {
-                    'nictype': 'bridged',
-                    'hwaddr': vif['address'],
-                    'parent': network_config['bridge'],
-                    'type': 'nic',
-                }
-            }
-        else:
-            config_update = {
-                net_device: {
-                    'nictype': 'p2p',
-                    'hwaddr': vif['address'],
-                    'type': 'nic',
-                }
+        net_device = lxd_vif.get_vif_devname(vif)
+        config_update = {
+            net_device: {
+                'nictype': 'physical',
+                'hwaddr': vif['address'],
+                'parent': lxd_vif.get_vif_internal_devname(vif),
+                'type': 'nic',
             }
+        }
 
         profile.devices.update(config_update)
         profile.save(wait=True)
 
     def detach_interface(self, context, instance, vif):
-        self.vif_driver.unplug(instance, vif)
-
         profile = self.client.profiles.get(instance.name)
-        to_remove = None
-        for key, val in profile.devices.items():
-            if val.get('hwaddr') == vif['address']:
-                to_remove = key
-                break
-        del profile.devices[to_remove]
-        profile.save(wait=True)
+        devname = lxd_vif.get_vif_devname(vif)
+
+        # NOTE(jamespage): Attempt to remove device using
+        #                  new style tap naming
+        if devname in profile.devices:
+            del profile.devices[devname]
+            profile.save(wait=True)
+        else:
+            # NOTE(jamespage): For upgrades, scan devices
+            #                  and attempt to identify
+            #                  using mac address as the
+            #                  device will *not* have a
+            #                  consistent name
+            for key, val in profile.devices.items():
+                if val.get('hwaddr') == vif['address']:
+                    del profile.devices[key]
+                    profile.save(wait=True)
+                    break
+
+        self.vif_driver.unplug(instance, vif)
 
     def migrate_disk_and_power_off(
             self, context, instance, dest, _flavor, network_info,
@@ -710,6 +869,23 @@ class LXDDriver(driver.ComputeDriver):
         """
         self.unpause(instance)
 
+    def resume_state_on_host_boot(self, context, instance, network_info,
+                                  block_device_info=None):
+        """resume guest state when a host is booted."""
+        try:
+            state = self.get_info(instance).state
+            ignored_states = (power_state.RUNNING,
+                              power_state.SUSPENDED,
+                              power_state.NOSTATE,
+                              power_state.PAUSED)
+
+            if state in ignored_states:
+                return
+
+            self.power_on(context, instance, network_info, block_device_info)
+        except (exception.InternalError, exception.InstanceNotFound):
+            pass
+
     def rescue(self, context, instance, network_info, image_meta,
                rescue_password):
         """Rescue a LXD container.
@@ -829,14 +1005,26 @@ class LXDDriver(driver.ComputeDriver):
                  int(cpu_topology['threads']))
 
         local_memory_info = _get_ram_usage()
-        local_disk_info = _get_fs_info(CONF.lxd.root_dir)
+
+        lxd_config = self.client.host_info
+
+        # NOTE(jamespage): ZFS storage report is very LXD 2.0.x
+        #                  centric and will need to be updated
+        #                  to support LXD storage pools
+        storage_driver = lxd_config['environment']['storage']
+        if storage_driver == 'zfs':
+            local_disk_info = _get_zpool_info(
+                lxd_config['config']['storage.zfs_pool_name']
+            )
+        else:
+            local_disk_info = _get_fs_info(CONF.lxd.root_dir)
 
         data = {
             'vcpus': vcpus,
-            'memory_mb': local_memory_info['total'] / units.Mi,
-            'memory_mb_used': local_memory_info['used'] / units.Mi,
-            'local_gb': local_disk_info['total'] / units.Gi,
-            'local_gb_used': local_disk_info['used'] / units.Gi,
+            'memory_mb': local_memory_info['total'] // units.Mi,
+            'memory_mb_used': local_memory_info['used'] // units.Mi,
+            'local_gb': local_disk_info['total'] // units.Gi,
+            'local_gb_used': local_disk_info['used'] // units.Gi,
             'vcpus_used': 0,
             'hypervisor_type': 'lxd',
             'hypervisor_version': '011',
@@ -1030,8 +1218,8 @@ class LXDDriver(driver.ComputeDriver):
                 cdb.make_drive(iso_path)
             except processutils.ProcessExecutionError as e:
                 with excutils.save_and_reraise_exception():
-                    LOG.error(_LE('Creating config drive failed with '
-                                  'error: %s'),
+                    LOG.error('Creating config drive failed with '
+                              'error: %s',
                               e, instance=instance)
 
         configdrive_dir = os.path.join(
